from collections import defaultdict


class Graph(object):
    def __init__(self):
        self.nodes = dict()
        self.edges = defaultdict(set)

    def add_connection(self, connection):
        for n in connection.first:
            self.nodes[n] = n
        for n in connection.second:
            self.nodes[n] = n
        
        for a in connection.first:
            for b in connection.second:
                if connection.op == "->" or connection.op == "--":
                    self.edges[a].add(b)
                if connection.op == "<-" or connection.op == "--":
                    self.edges[b].add(a)

    def direct_successors_of(self, node):
        return self.edges[node]

    def successors_of(self, node):
        queue = [node]
        s = set()
        while queue:
            n = queue.pop()
            for nn in self.direct_successors_of(n):
                if nn not in s:
                    s.add(nn)
                    queue.append(nn)
            
        return s

    def direct_predecessors_of(self, node):
        for n in self.nodes:
            if node in self.edges[n]:
                yield n

    def predecessors_of(self, node):
        """
            Not an efficient algorithm.
            Not clear if we care about this - do we
            want to use some other graph library?
        """
        for n in self.nodes: 
            if node in self.successors_of(n):
                yield n

    def add_description(self, node, description):
        self.nodes[node] = description

    def roots(self):
        s = set(self.nodes)
        for n in self.nodes:
            for m in self.edges[n]:
                if m in s:
                    s.remove(m)
        return s

    def topological_sort(self):
        total = 0
        count = defaultdict(int) 
        for n in self.nodes:
            for m in self.edges[n]:
                count[m] = count[m] + 1
                total = total + 1

        queue = list(self.roots())

        while queue:
            n = queue.pop()
            yield n
            for m in self.edges[n]:
                count[m] = count[m] - 1
                total = total - 1
                if count[m] == 0:
                    queue.append(m)

        if total > 0:
            raise Exception("Tried to topological sort a graph with a cycle.")


def make_graph(spec):
    connection_list, node_list = spec
    graph = Graph()
    for connections in connection_list:
        for connection in connections:
            graph.add_connection(connection)
    for node, description in node_list:
        graph.add_description(node, description)
    return graph



import re

import ply.lex as lex
import ply.yacc as yacc

from .graph import make_graph
from .internal import Connection
from .preprocess import preprocess


tokens = (
    'NODENAME',
    'QUOTED',
    'LEFT_CONNECT',
    'RIGHT_CONNECT',
    'BOTH_CONNECT',
    'COMMA',
    'NEWLINE',
    'EQUALS',
    'BACKTICK',
    'TEXT',
)

t_NODENAME = r'[a-zA-Z_:][a-zA-Z_:0-9]*'
t_QUOTED = r'"[a-zA-Z_:][a-zA-Z_:0-9 \t]*"'
t_LEFT_CONNECT = r'->'
t_RIGHT_CONNECT = r'<-'
t_BOTH_CONNECT = r'--'
t_COMMA = r','
t_NEWLINE = r'\n'
t_EQUALS = r'='
t_TEXT = r'`[^`]*`'

t_ignore = " \t"

def t_error(t):
    print("Illegal character '%s'" % (t.value[0], ))
    t.lexer.skip(1)

lexer = lex.lex()


def tokenize(data):
    lexer.input(data)
    while True:
        tok = lexer.token()
        if not tok:
            break
        yield tok


def p_edges_and_nodes_base(p):
    'edges_and_nodes : '
    p[0] = ([], [])


def p_edges_and_nodes_edges(p):
    'edges_and_nodes : edges_and_nodes connection_list'
    p[0] = (p[1][0] + p[2], p[1][1])


def p_edges_and_nodes_nodes(p):
    'edges_and_nodes : edges_and_nodes node_spec'
    p[0] = (p[1][0], p[1][1] + [p[2]])


def p_node_name_nodename(p):
    'node_name : NODENAME'
    p[0] = p[1]


def p_node_name_quoted(p):
    'node_name : QUOTED'
    text = p[1][1:-1]
    p[0] = text


def p_node_spec(p):
    'node_spec : node_name EQUALS TEXT NEWLINE'
    text = p[3][1:-1]
    p[0] = (p[1], text)


def p_nodelist_node(p):
    'nodelist : node_name'
    p[0] = [p[1]]


def p_nodelist_comma(p):
    'nodelist : nodelist COMMA node_name'
    p[0] = p[1] + [p[3]]


def p_right_connect(p):
    'connect : RIGHT_CONNECT'
    p[0] = p[1]


def p_left_connect(p):
    'connect : LEFT_CONNECT'
    p[0] = p[1]


def p_both_connect(p):
    'connect : BOTH_CONNECT'
    p[0] = p[1]


def p_nodelist_connection(p):
    'connections : nodelist connect nodelist'
    p[0] = [Connection(p[1], p[3], p[2])]


def p_nodelist_further_connections(p):
    'connections : connections connect nodelist'
    p[0] = p[1] + [Connection(p[1][-1].second, p[3], p[2])]


def p_connectionlist_empty_line(p):
    'connection_list : NEWLINE'
    p[0] = []


def p_connectionlist_line(p):
    'connection_list : connections NEWLINE'
    p[0] = [p[1]]


def p_error(p):
    print("Syntax error in input: %s" % (p, ))


parser = yacc.yacc()


def parse(data):
    clean_data = preprocess(data)
    result = parser.parse(clean_data)
    return make_graph(result)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['virgo']

package_data = \
{'': ['*']}

install_requires = \
['ply>=3.11,<4.0']

setup_kwargs = {
    'name': 'pyvirgo',
    'version': '0.2.0',
    'description': 'A Python package for the Virgo language',
    'long_description': '# PyVirgo\n\nThis is a Python implementation of Virgo, the graph declarative language.\n\nYou can find details of Virgo with the Go implementation at https://github.com/r2d4/virgo\n\nVirgo is designed so that we can express graphs in a config file. These could be dependency graphs, for example of a build process, or any other graph structure.\n\nTo invoke PyVirgo:\n>>> import virgo\n>>> g = virgo.loads("a -> b, c <- d")\n>>> g       # doctest: +ELLIPSIS\n<virgo.graph.Graph object at ...>\n>>> sorted(list(g.direct_successors_of("a")))\n[\'b\', \'c\']\n\nIt\'s more likely that we will want to load a graph from a file:\n>>> g2 = virgo.load("test/files/make.vgo")\n>>> g2      # doctest: +ELLIPSIS\n<virgo.graph.Graph object at ...>\n>>> g2.direct_successors_of("src files")\n{\'test\'}\n\nWe can access the \'node data\' for each node, by identifier.\n>>> g2.nodes["src files"]\n\'go build ./...\'\n',
    'author': 'Jack Grahl',
    'author_email': 'jack.grahl@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/jwg4/pyvirgo',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import re


MULTILINE_COMMENT_REGEX = re.compile(r"[/][*][^*]*([*][^/][^*]*)*[*][/]")
CONTINUATION_REGEX = re.compile(r"[|]\n")
COMMENT_REGEX = re.compile(r"[/]{2}.*$")

def _gen_preprocess(data):
    data = MULTILINE_COMMENT_REGEX.sub("", data)
    data = CONTINUATION_REGEX.sub("", data)
    for line in data.split("\n"):
        yield COMMENT_REGEX.sub("", line)        
    yield ""


def preprocess(data):
    return "\n".join(_gen_preprocess(data))
